import math
from . import *

C_MAIN = std_core(20, 8, 
                 ["LED", "SDAT", "SCTL",
                  "FCI0", "FCI1",
                  "MUH", "MUL", "RND",
                  "TTL", "GPIO", "ATRG",
                  "ECTR", "CTRX"],
                  {"GPIO": ["DIR", "INV", "POS", "NEG", "TRG"],
                   "CTRX": [f"C{i:X}" for i in range(32)]},
                 131072, 131072)

def CORE_RWG():
    regs = ["LED", "SDAT", "SCTL", "FCI0", "FCI1",
            "MUH", "MUL", "RND", "TTL", "GPIO", 
            "ECTR", "CTRX", "CMON", "CDDS", "CSBG",
            "MSBG", "DLAY"]
    for w in ["F", "P", "A"]:
        for s in range(64):
            regs += [f"{w}{s:02X}"]
    msbg = []
    for d in range(4):
        msbg += [f"MX{d}L", f"MX{d}H"]
    msbg += [f"SCA{i}" for i in range(4)] + ["OFSC"]
    sub = {"GPIO": ["DIR", "INV", "POS", "NEG", "TRG"],
           "CTRX": [f"C{i:X}" for i in range(4)],
           "CMON": [f"D{i:X}" for i in range(4)],
           "MSBG": msbg}
    return std_core(20, 0, regs, sub, 131072, 131072)

C_RWG = CORE_RWG()

A0 = (0, 0, 0, 0)
A1 = (1, 1, 1, 1)

@multi(asm)
def init(ofsc=19):
    """
    
    Reset & initialize the registers & peripherals.

    Returns
    -------
    None.

    """
    timer()
    ldi('led',0)
    for s in range(64):
        mov(f'F{s:02X}','nul',B=0)
        mov(f'P{s:02X}','nul',B=0)
        mov(f'A{s:02X}','nul',B=0)
    sbg.signal()
    sbg.ctrl(sfq=A1, cph=A1, sam=A1, pud=A1)
    signal(prof=A0)
    ctrl(rst=A1)
    delay(0, 0, 90, 57)
    sel('msbg','ofsc')
    ldi('msbg',ofsc,B=0)
    for chn in range(4):
        regwrite(chn, 0x1, [0x01, 0x01, 0x00, 0x20])
        regwrite(chn, 0x2, [0x00, 0x00, 0xC0, 0x00])
        regwrite(chn, 0x3, [0x00, 0x00, 0x00, 0xFF])
        regwrite(chn, 0xA, [0x18, 0x00, 0x00, 0x00])
        ducwrite(chn, 0, 2, 1, 0, 0, 0, 0)
        regwrite(chn, 0x0, [0x00, 0x60, 0x20, 0x02])
        sbg.mux(chn, 0xFFFF << (chn*16), 0)

@multi(asm)
def sync():
    """
    
    Synchronize the datalink between DDS and FPGA.

    Returns
    -------
    None.

    """
    timer(150)
    sbg.signal()
    sbg.ctrl(pud=(1, 1, 1, 1))
    timer(150)
    ctrl(synrst=1)

@multi(asm)
def spi_send(slv_adr, dst_reg, adr_len, dat_len, clk_div, ltn=0, wait=True):
    val = bit_concat((clk_div, 8), (dat_len, 4), (adr_len, 1),
                     (slv_adr, 3), (ltn, 4), (dst_reg, 12))
    ldi('sctl', val)
    if wait:
        sleep((clk_div + 1) * (16 * (dat_len + adr_len + 2)))
         
_prof = [0] * 4
    
@multi(asm)
def signal(iou=A0, mrk=A0, prof=(None,None,None,None),load=True):
    global _prof
    p = [0] * 4
    tmp = [0] * 4
    for i in range(4):
        p[i] = _prof[i] if prof[i] is None else prof[i]
        tmp[i] = bit_concat((iou[i], 1), (mrk[i], 1),
                            (p[i], 3))
    _prof = p
    val = bit_concat((0, 12),
                        (tmp[3], 5), (tmp[2], 5),
                        (tmp[1], 5), (tmp[0], 5))
    if load:
        ldl('cdds',val)
    else:
        return val

@multi(asm)
def ctrl(rst=A0, iorst=A0, synrst=0, load=True):
    tmp = [0] * 4
    for i in range(4):
        tmp[i] = bit_concat((rst[i], 1), (iorst[i], 1))
    val = bit_concat((0, 3), (synrst, 1),
                        (tmp[3], 2), (tmp[2], 2),
                        (tmp[1], 2), (tmp[0], 2), 
                        (0, 20))
    if load:
        ldh('cdds',val)
    else:
        return val

@multi(asm)
def delay(prof=0, mrk_sgt=0, mrk_sbg=90, iou_sbg=57):
    val = bit_concat((prof, 8), (mrk_sgt, 8), (mrk_sbg, 8), (iou_sbg, 8))
    ldi('dlay',val)

def sig_msk(chn, val=1, oth=0):
    ret = [oth] * 4
    ret[chn] = val
    return ret

RegLen = [4, 4, 4, 4, 4, 6, 6, 4,
            2, 4, 4, 8, 8, 4, 8, 8,
            8, 8, 8, 8, 8, 8, 4, 0, 2, 2]
CDiv_DDSW = 1
CDiv_DDSR = 5
LTN_DDSR = 1

@multi(asm)
def regwrite(chn, reg, dat, clk=None, wait=True, ioupd=True):
    tln = len(dat)
    dat = dat + ([0] * (-tln % 4))
    ln = len(dat) // 4
    ins = [0] * ln
    for i in range(ln):
        ins[i] = bit_concat((dat[i*4], 8), (dat[i*4+1], 8),
                            (dat[i*4+2], 8), (dat[i*4+3], 8))
    for i in range(ln):
        ldi('sdat',ins[ln-i-1],B=1)
    spi_send(chn+1, reg, 0, tln, clk or CDiv_DDSW, ltn=0, wait=wait)
    if ioupd:
        signal(iou=sig_msk(chn))

@multi(asm)
def profwrite(chn, prof, frq, amp, pha, wait=True, ioupd=True):
    ftw = round((frq / 1200) * (2 ** 32))
    ftw = ftw.to_bytes(4, "big")
    asf = round(amp * 16383).to_bytes(2, "big")
    phw = round(pha * 65536).to_bytes(2, "big")
    regwrite(chn, prof + 14, list(asf + phw + ftw), wait=wait, ioupd=ioupd)

@multi(asm)
def ducwrite(chn, prof, ccir, s_inv, i_cci, frq, amp, pha, wait=True, ioupd=True):
    ctr = bit_concat((ccir, 6), (s_inv, 1), (i_cci, 1)).to_bytes(1, "big")
    ftw = round((frq / 1200) * (2 ** 32)).to_bytes(4, "big")
    asf = round(amp * 255).to_bytes(1, "big")
    phw = round(pha * 65536).to_bytes(2, "big")
    regwrite(chn, prof + 14, list(ctr + asf + phw + ftw), wait=wait, ioupd=ioupd)

@multi(asm)
def carrier(chn, frq, amp=1.0, pha=0.0, upd=False):
    ducwrite(chn, 0, 2, 1, 0, frq, amp, pha, wait=True, ioupd=upd)

@multi(asm)
def wait_master():
    timer()
    smk('err','!nul','0x1.0',B=0)
    nop(4)

@multi(asm)
def trig_slave(slave, offset=476):
    """
    
    Trigger the selected slave boards. Used in pair with <Wait_Master>.

    Parameters
    ----------
    slave : <list>
        List of addresses of the selected slave boards.
    offset : <int> or <float>, optional
        Delay after sending the trigger. Use it to align the behavior
        of the main board with the slave boards.
        
        <int>: the duration is in unit of cycles (3.33ns).
        
        <float>: the duration is in unit of micro-second.
        
        The default is 476. With this value, the GPIO output is aligned with
        MARK output of the RWG boards with RTLink baud-rate = 2.

    Returns
    -------
    None.

    """
    bce = sum([1 << (i-1) for i in slave])
    if isinstance(offset, float):
        offset = round(offset * 300)
    timer(offset)
    set_rtcf(hop=0, typ=1, bce=bce, bcs=0)
    nop(2)
    with asm:
        ins = smk('err','nul','0x1.0')
    ldi('rtbc',ins)
    set_rtcf(typ=0)
    
@multi(asm)
def cfg_gpio(dir=None, inv=None, pos=None, neg=None, trg=None):
    """
    
    Configurations of GPIO ports. Set the direction, inversion, sensitivity
    of each port and whether a port is used as trigger. The parameters are 
    32-bit integers with each bit corresponds to a port. The power-up defaults
    are 0.
    
    inv, pos, neg and trg take effect only for input ports. The input level
    is first optionally inverted, then the sensitive edges are converted to
    pulses, and finally used for triggering and counting.
    
    If neither pos nor neg is asserted for a port,
    the port is high-level sensitive.
    
    If more than one ports are used as trigger, any one can fire the trigger.
    
    Parameters
    ----------
    dir : <int>, optional
        Direction of the GPIO ports, 1 for input, 0 for output.
    inv : <int>, optional
        Logic invert flag, 1 for invert.
    pos : <int>, optional
        Positive-edge sensitive flag, 1 for sensitive.
    neg : <int>, optional
        Negative-edge sensitive flag, 1 for sensitive.
    trg : <int>, optional
        Trigger enable flag, 1 for used as trigger.

    Returns
    -------
    None.

    """
    if dir is not None:
        sel('gpio','dir')
        mov('gpio',dir)
    if inv is not None:
        sel('gpio','inv')
        mov('gpio',inv)
    if pos is not None:
        sel('gpio','pos')
        mov('gpio',pos)
    if neg is not None:
        sel('gpio','neg')
        mov('gpio',neg)
    if trg is not None:
        sel('gpio','trg')
        mov('gpio',trg)
        
def _get_scale(val, wid):
    orig = val
    tmp = max(-val - 1 if val < 0 else val, 1)
    scl = max(round(math.log2(tmp)) + 2 - wid, 0)
    val = val >> scl
    if scl >= 16 or (orig != 0 and val == 0) or not (-2048 <= val < 2048):
        raise
    return val, scl

def _adr_sb(idx):
    return (None if len(idx) < 2 else idx[0]),idx[-1]     

class SBG(meta):
    @staticmethod                        
    def signal(txe=A0, mrk=A0):
        tmp = [0] * 4
        for i in range(4):
            tmp[i] = bit_concat((0, 1), (txe[i], 1), (mrk[i], 1))
        val = bit_concat((tmp[3], 3), (tmp[2], 3),
                         (tmp[1], 3), (tmp[0], 3), (0, 20))
        ldh('csbg',val)

    @staticmethod
    def ctrl(iou=A0, sfq=A0, cph=A0, sam=A0, pud=A0):
        tmp = [0] * 4
        for i in range(4):
            tmp[i] = bit_concat((iou[i], 1), (sfq[i], 1), (cph[i], 1),
                                (sam[i], 1), (pud[i], 1))
        val = bit_concat((0, 12),
                         (tmp[3], 5), (tmp[2], 5),
                         (tmp[1], 5), (tmp[0], 5))
        ldl('csbg',val)
    
    @staticmethod
    def play(dur, phase_origin=False, txe=A1, mrk=A0, sfq=A0, sam=A0, strict=True):
        """
        
        Apply the SBG parameter changes and start generating waves.

        Parameters
        ----------
        dur : <int> or <float>
            Duration of this PlayWave stage.

            <int>: the duration is in unit of cycles (3.33ns).
            
            <float>: the duration is in unit of micro-second.
            
        phase_origin : <tuple> or <bool>, optional
            Whether to clear phase accumulators of the carrier and sidebands.
            Should be 2-tuple of 4-tuples, e.g. ((0, 0, 0, 0), (1, 1, 1, 1)),
              with the first tuple corresponding to the carrier.
            True is the same as ((1, 1, 1, 1), (1, 1, 1, 1));
              False is the same as ((0, 0, 0, 0), (0, 0, 0, 0)).
            The default is False.
        txe : tuple of 4, optional
            Parallel modulation enable flag of each RF port, aligned with RF waveform output.
            The default is (1, 1, 1, 1).
        mrk : tuple of 4, optional
            Output state of each GPIO port, aligned with RF waveform output.
            The default is (0, 0, 0, 0).
        sfq : tuple of 4, optional
            Frequency update flag of each RF port.
            1 for clear frequency ramp accumulator and set new sideband frequency.
            Affect all SBGs assigned to the corresponding RF channel.
            The default is (0, 0, 0, 0).
        sam : tuple of 4, optional
            Amplitude update flag of each RF port.
            1 for clear amplitude ramp accumulator and set new sideband amplitude.
            Affect all SBGs assigned to the corresponding RF channel.
            The default is (0, 0, 0, 0).

        Returns
        -------
        None.

        """
        if phase_origin is True:
            car = A1
            cph = A1
        elif phase_origin is False:
            car = A0
            cph = A0
        else:
            car, cph = phase_origin
        if type(dur) is float:
            dur = round(dur * 300)
        if type(dur) is int:
            dur = dur & -2
        if type(mrk) is int:
            mrk = bit_split(mrk,(1,1,1,1))
        timer(dur,strict=strict)
        sbg.signal(txe=A1, mrk=mrk)
        sbg.ctrl(iou=car, sfq=sfq, cph=cph, sam=sam, pud=A1)
        
    @staticmethod
    def mux(chn, ena=None, scale=None):
        """
        
        Configure the association of SBGs to the RF channels.

        Parameters
        ----------
        chn : <int>, 0-3
            The RF channel number to be configured.
        ena : <int>
            64 bit integer, each bit corresponds to a SBG. 1 for associating this SBG to the RF channel.
        scale : <int>
            SBG signal attenuation level for this RF channel.
            The signals of the SBGs associated to a channel are first summed 
            then shifted right <scale> bits before output to dds.
        
        NOTE: By default, SBG 0-15 are with RF channel 0, 16-31 with channel 1, etc. And scale is 0.

        Returns
        -------
        None.

        """
        if ena is not None:
            sel('msbg', f"MX{chn}L")
            ldi('msbg', ena)
            sel('msbg', f"MX{chn}H")
            ldi('msbg', ena >> 32)
        if scale is not None:
            sel('msbg', f"SCA{chn}")
            ldi('msbg', scale)

    @staticmethod
    def ftw(sb,frq,high_res=True):
        '''
        frq : <float>, -60.0 - +60.0
        Sideband frequency offset in MHz.
        '''
        ftw = round((frq / 150.0) * 0x1_0000_0000)
        ftw_crs = (ftw >> 12) & 0xFFFFF
        ftw_fin = ftw & 0xFFF
        ftw = bit_concat((ftw_fin, 12), (ftw_crs, 20))
        if sb is None:
            return ftw
        r_f = f"F{sb:02X}"
        ldl(r_f,ftw,B=0)
        if high_res:
            ldh(r_f,ftw,B=0)

    @staticmethod 
    def asf(sb,am_p,am_n=0,high_res=True):
        '''
        am_p : <float>, 0.0 - 1.0
            Positive sideband amplitude.
        am_n : <float>, 0.0 - 1.0
            Negative sideband amplitude. NOTE: abs(am_p ± am_n) <= 1
        '''
        asp = round(am_p * 0x7FFF)
        asp_crs = (asp >> 6) & 0x3FF
        asp_fin = asp & 0x3F
        asn = round(am_n * 0x7FFF)
        asn_crs = (asn >> 6) & 0x3FF
        asn_fin = asn & 0x3F
        asf = bit_concat((asn_fin, 6), (asp_fin, 6), 
                         (asn_crs, 10), (asp_crs, 10))
        if sb is None:
            return asf
        r_a = f"A{sb:02X}"
        ldl(r_a,asf,B=0)
        if high_res:
            ldh(r_a,asf,B=0)

    @staticmethod
    def pof(sb,pha,org=0,high_res=True):
        '''
        pha : <float>, 0.0 - 1.0
        Sideband phase offset, in unit of 2PI.
        '''
        pof = round(pha * 0x1_00000 + (org >> 13)) & 0xFFFFF
        if sb is None:
            return pof
        r_p = f"P{sb:02X}"
        ldl(r_p,pof,B=0)
        if high_res:
            ldh(r_p,0,B=0)    

    @staticmethod
    def ramp(sb,fsi=None,ar_p=None,ar_n=None):
        '''
        fsi : <float>, signed, 1e-5 - 350.0, optional
        Frequency ramp rate, in unit of MHz/us.
        Setting to None disables frequency ramp.
        ar_p : <float>, signed, 0.005 - 18.0, optional
            Positive sideband amplitude ramp rate, in unit of 1/us.
        ar_n : <float>, signed, 0.005 - 18.0, optional
            Negative sideband amplitude ramp rate, in unit of 1/us.
            Setting ar_p or ar_n to None disables amplitude ramp.
        '''
        if fsi is not None:
            fsi = round((fsi / 22500.0) * 0x1_0000_0000)
            fsi, fsf = _get_scale(fsi, 12)
            fsi <<= 20
            fse = 1
        else:
            fsf, fse = 0, 0
        if ar_p is not None or ar_n is not None:
            if ar_p is None:
                ar_p = 0
            if ar_n is None:
                ar_n = 0
            arp = round(ar_p * 0x7FFF / 150.0)
            arn = round(ar_n * 0x7FFF / 150.0)
            arp, afp = _get_scale(arp, 6)
            arn, afn = _get_scale(arn, 6)
            arx = bit_concat((arn, 6), (arp, 6)) << 20
            are = 1
        else:
            arx = None
            afp, afn, are = 0, 0, 0
        scn = bit_concat((fse, 1), (are, 1), (fsf, 4), (afn, 3), (afp, 3)) << 20
        if sb is None:
            return fsi,arx,scn
        r_f = f"F{sb:02X}"
        r_p = f"P{sb:02X}"
        r_a = f"A{sb:02X}"
        if fsi is not None:
            ldh(r_f,fsi,B=0)
        if arx is not None:
            ldh(r_a,arx,B=0)
        ldh(r_p,scn,B=0)       
    
    def f(self,frq,high_res=True):
        adr,sb = _adr_sb(self.__meta__)
        self.ftw[adr](sb,frq,high_res)
        return self
    
    def a(self,am_p,am_n=0,high_res=True):
        adr,sb = _adr_sb(self.__meta__)
        self.asf[adr](sb,am_p,am_n,high_res)
        return self
     
    def p(self,pha,org=0,high_res=True):
        adr,sb = _adr_sb(self.__meta__)
        self.pof[adr](sb,pha,org,high_res)
        return self
    
    def r(self,fsi=None,ar_p=None,ar_n=0):
        adr,sb = _adr_sb(self.__meta__)
        self.ramp[adr](sb,fsi,ar_p,ar_n)
        return self

SBG.ctrl = multi(asm,SBG.ctrl)
SBG.signal = multi(asm,SBG.signal)
SBG.play = multi(asm,SBG.play)
SBG.mux = multi(asm,SBG.mux)
SBG.ftw = multi(asm,SBG.ftw)
SBG.asf = multi(asm,SBG.asf)
SBG.pof = multi(asm,SBG.pof)
SBG.ramp = multi(asm,SBG.ramp)

sbg = SBG()

@multi(asm)
def ttl_level(ttl):
    mov('ttl',ttl,B=0)

@multi(asm)
def ttl_stage(dur,ttl):
    if type(dur) is float:
        dur = round(dur * 300)
    if type(dur) is int:
        dur = dur & -2
    timer(dur)
    ttl_level(ttl)
    
@multi(asm)
def save_count(ctr, ext=False):
    """
    
    Save the counts of the selected counters to data cache.
    The counts can only be accessed after deactivation.

    Parameters
    ----------
    ctr : <int>
        32-bit integer, each bit corresponds to a counter.
        1 for selecting this channel.

    Returns
    -------
    None.

    """
    sleep(120)
    ldi('dca',0)
    nop(4)
    mov('dca','dcd')
    for i in range(3 if ext else 4):
        if ctr & (1 << i) > 0:
            sel('ctrx',f"C{i:X}")
            add('dca','dca',1,B=0)
            mov('dcd','ctrx',B=0)
            nop(2)
    if ext and (ctr & 0b1000 > 0):
        sleep(24000)
        for i in range(4, 36):
            sel('ctrx',f"C{i:X}")
            add('dca','dca',1,B=0)
            mov('dcd','ctrx',B=0)
            nop(2)
    nop(2)
    mov('tr0','dca',B=0)
    ldi('dca',0,B=0)
    nop(2)
    mov('dcd','tr0')
    
if __name__ == '__main__':
    with asm as f:
        asm.multi=[1, 2, 3]
        setup(C_RWG)
        #init()
        #sbg.ftw(0x00,10)
        #sbg.asf[1](0x00,0.5)
        #sbg.pof[2,3](0x00,0)
        sbg[0x00].f(10).a(0.5)
        sbg[1,2][0x00].p(0)
        finish()
    print(f['1'][:],f['2'][:],f['3'][:])


import pkgutil
__path__ = pkgutil.extend_path(__path__,__name__)
__path__.reverse()

from contextlib import contextmanager
from .. import *

def bit_concat(*args):
    """

    Bit fields concatenation.

    Parameters
    ----------
    *args : 2-tuples as (V, W)
        Each argument corresponds to a bit field with value V and width W.
        Bit fields are concatenated from high to low.
        NOTE: V will be cropped if not less than (2 ** W)

    Returns
    -------
    dat : integer
        Concatenated value.

    Example
    -------
    bit_concat((0b11, 2), (0b10, 3), (0b11, 1)) == 0b110101

    """
    dat = 0
    ln = len(args)
    for i in range(ln - 1):
        dat |= args[i][0] & ((1 << args[i][1]) - 1)
        dat <<= args[i + 1][1]
    dat |= args[ln - 1][0] & ((1 << args[ln - 1][1]) - 1)
    return dat

def bit_split(dat, wids):
    """

    Split a number into bit fields.

    Parameters
    ----------
    dat : integer
        Number to be splitted.
    wids : list OR tuple of integers
        Width of each bit field.

    Returns
    -------
    bf : list of integers
        Values of bit fields.

    Example
    -------
    bit_split(0b110101, (2, 3, 1)) == [0b11, 0b10, 0b1]

    """
    bf = []
    for i in range(1, len(wids)):
        bf += [dat & ((1<<wids[-i])-1)]
        dat = dat >> wids[-i]
    bf += [dat & ((1<<wids[0])-1)]
    bf.reverse()
    return bf
    
class base_core:
    """

    Base class for RTMQ core structure,
    defines constants and registers common to the RTMQ framework.

    Constants
    ----------
    OPC   : Type-A instruction opcodes.
    MGW   : Magic word of RTLink frame.
    RSV   : Reserved registers (core registers & system peripherals).
    T_ERR : Error message.

    Attributes
    ----------
    REG     : List of register names.
    SBF     : Dict of register sub-file entries.
    CAP_ICH : Maximum capacity of instruction cache.
    CAP_DCH : Maximum capacity of data cache.

    """
    OPC = {"ADD": "AA", "AND": "LL", "XOR": "LL",
           "CLU": "AA", "CLS": "AA", "CEQ": "LA",
           "SGN": "AA", "SNE": "LA", "SMK": "LL", "MOV": "RR",
           "SLL": "LA", "SLA": "LA", "SLC": "LA"}
    MGW = 0xA
    MGW_ERR = 0x5
    RSV = ["NUL", "PTR", "WCK", "TIM", "TRG", "ERR",
           "ICA", "ICD", "DCA", "DCD", "RTS", "DTS",
           "RTCF", "RTBC", "RTDA"]
    T_ERR = ["Core suspended",
             "ALU register write conflict",
             "Timeout exceeded",
             "Trigger misaligned"]

    def __init__(self, regs, sbfs, cap_ich, cap_dch):
        """

        Parameters
        ----------
        regs : list of strings
            List of all other registers.
        sbfs : dict of lists of strings
            The key of each item is the entry register name, 
            and the value is a list of names of the registers in the sub-file.
        cap_ich : integer
            Maximum capacity of instruction cache.
        cap_dch : integer
            Maximum capacity of data cache.

        Returns
        -------
        None.

        """
        self.REG = self.RSV + regs
        self.alu_opc = dict()
        opc = list(self.OPC.keys())
        for i in range(len(opc)):
            self.alu_opc[opc[i]] = i
        self.regs = dict()
        for i in range(len(self.REG)):
            self.regs[self.REG[i]] = i
        self.SBF = sbfs
        self.sbfs = dict()
        for k, v in self.SBF.items():
            self.sbfs[k] = dict()
            for i in range(len(v)):
                self.sbfs[k][v[i]] = i
        self.CAP_ICH = cap_ich
        self.CAP_DCH = cap_dch

class std_core(base_core):
    """

    Standard RTMQ core structure
    with basic computation and flow control capability.

    Constants
    ----------
    RSV_GPR : Assembler reserved general purpose registers. \n
    N_PLM   : Cache pipeline latency. \n
    N_PLI   : Type-I instruction pipeline latency. \n
    N_PLA   : Type-A instruction pipeline latency. \n
    N_PLS   : Stack pop pipeline latency. \n
    T_ERR   : Error message.


    """

    RSV_GPR = ["TR0", "TR1", "NOW"]

    N_PLM = 3
    N_PLI = 1
    N_PLA = 4
    N_PLS = 4

    T_ERR = ["PTR out of bound",
             "ICA out of bound",
             "DCA out of bound",
             "RTS overflow",
             "RTS underflow",
             "DTS overflow",
             "DTS underflow",
             "RTLink Rx FIFO overflow",
             "RTLink ExtUART Tx conflict",
             "Error occured in slave nodes"]

    def __init__(self, n_gpr, n_slv, regs, sbfs, cap_ich, cap_dch):
        """

        Parameters
        ----------
        n_gpr : integer
            Number of non-reserved general-purpose registers.
        n_slv : integer
            Number of RTLink slaves (within 0-15).
        regs : list of strings
            List of all other registers.
        sbfs : dict of lists of strings
            The key of each item is the entry register name, 
            and the value is a list of names of the registers in the sub-file.
        cap_ich : integer
            Maximum capacity of instruction cache.
        cap_dch : integer
            Maximum capacity of data cache.

        Returns
        -------
        None.

        """
        self.n_gpr = n_gpr
        treg = self.RSV_GPR.copy()
        for i in range(n_gpr):
            treg += [f"R{i:X}"]
        treg += regs
        tsbf = {"RTDA": []}
        for i in range(n_slv + 1):
            tsbf["RTDA"] += [f"C{i:X}"]
        for k, v in sbfs.items():
            tsbf[k] = v
        super().__init__(treg, tsbf, cap_ich, cap_dch)
        self.T_ERR = super().T_ERR + self.T_ERR

# Global instance used to implement basic operations
C_BASE = base_core([], {"RTDA": ["C0"]}, 65536, 16384)
C_STD = std_core(20, 0, [], {}, 65536, 16384)

def source(code, core=None, line=True):
    """

    Convert machine code program back to assembly program.

    Parameters
    ----------
    code : list of integers
        Program in machine code.
    core : <CoreBase>, optional
        Destination core structure descriptor. The default is None.

    Returns
    -------
    prg : multi-line string
        Assembly program.

    """
    if core is None:
        core = getattr(code,'core',asm.core)
    if type(code) is int:
        ith, itl, h, f, opc, rd, i0, i1, n0, n1, rx = \
            bit_split(code, (1, 1, 1, 1, 4, 8, 1, 1, 1, 1, 12))
        cod = ("LDH", "LDL")[itl] if ith == 0 else list(core.OPC.keys())[opc]
        if cod == "LDH":
            if rd == 0:
                cod = "NOP"
            elif opc >= 0x8:
                cod = "SEL"
        hld = ("-", "H")[h]
        fct = ("-", "F")[f]
        head = f"{cod} {hld} {fct}"
        rds = f"{core.REG[rd]:4}"
        if ith == 0:
            imm = bit_concat((opc, 4), (i0, 1), (i1, 1),
                             (n0, 1), (n1, 1), (rx, 12))
            if itl == 0:
                if opc >= 0x8:
                    s = bit_concat((opc, 3), (i0, 1), (i1, 1), (n0, 1), (n1, 1))
                    ims = core.SBF[core.REG[rd]][s]
                else:
                    ims = f"0x{imm:03X}_00000"
            else:
                ims = f"0x000_{imm:05X}"
            if cod == "NOP":
                instr = head
            else:
                instr = f"{head}  {rds}  {ims}"
        else:
            ird = (" ", "!")[itl]
            ir0 = (" ", "!")[n0]
            ir1 = (" ", "!")[n1]
            r0, r1 = (0, rx) if opc == 9 else bit_split(rx, (6, 6))
            r0l = bit_split(bit_concat((n0, 1), (r0, 6)), (3, 4))
            r0l = f" 0x{r0l[1]:X}.{r0l[0]}"
            r0i = f" 0x{r0:02X}" if n0 == 0 else f"-0x{0x40-r0:02X}"
            r0i = {"R": "", "A": r0i, "L": r0l}[core.OPC[cod][0]]
            r0s = f"{ir0}{core.REG[r0]}" if i0 == 0 else r0i
            r1l = bit_split(bit_concat((n1, 1), (r1, 6)), (3, 4))
            r1l = f" 0x{r1l[1]:X}.{r1l[0]}"
            r1i = f" 0x{r1:02X}" if n1 == 0 else f"-0x{0x40-r1:02X}"
            r1i = {"R": "", "A": r1i, "L": r1l}[core.OPC[cod][1]]
            r1s = f"{ir1}{core.REG[r1]}" if i1 == 0 else r1i
            if cod == "MOV":
                instr = f"{head} {ird}{rds} {r1s}"
            else:
                instr = f"{head} {ird}{rds} {r0s} {r1s}"
        return instr
    elif type(code) in (tuple,list,table):
        prg = [source(ins,core) for ins in code]
        if line:
            return '\n'.join([f'{i:05X}:\t{prg[i]}' for i in range(len(prg))])
        else:
            return '\n'.join(prg)
    return str(code)
        
'''
core : <CoreBase>, optional
    Destination core structure descriptor. The default is None.
label : dict, optional
    Dict of jump labels. The default is dict().
'''
asm = context(core=C_STD)

def label(tag,put=True):
    lbl = getattr(asm,'label',None)
    if lbl is None:
        lbl = {}
        asm.label = lbl
    pos = lbl.get(tag,None)
    if put is True or type(put) is int:
        if put is True:
            put = len(asm)
        if type(pos) is list:
            pos[0][0] = put
            for i in pos[1:]:
                asm[i] = asm[i]()
        lbl[tag] = put
    else:
        if type(pos) is int:
            return pos
        if pos is None:
            pos = [expr(put)]
            lbl[tag] = pos
        pos.append(len(asm))
        return pos[0]
        
def cnv_opd(opd, opc, field, typ):
    if type(opd) not in (int,float,str,core_reg):
        if "A" not in typ:
            raise
        return (label(id(opd),opd),1,0)
    opd = str(opd)
    try:
        if opd[0] == '#':
            if "A" not in typ:
                raise
            return (label(opd[1:],None), 1, 0)
        elif opd[0] == '-' or opd[0].isdigit():
            num = opd.split('.')
            if len(num) == 1:
                if "A" not in typ:
                    raise
                return (int(opd, 0) & 0xFFFF_FFFF, 1, 0)
            else:
                nib = int(num[0], 0)
                pos = int(num[1])
                if not ((0 <= nib <= 15) and (0 <= pos <= 7) and ("L" in typ)):
                    raise
                return ((pos << 4) | nib, 1, 0) 
        inv = 0
        while opd[0] == "!":
            opd = opd[1:]
            inv = 1 - inv
        if "R" not in typ:
            raise
        return (asm.core.regs[opd.upper()], 0, inv)
    except:
        raise SyntaxError(f"Invalid {field} for {opc}: '{opd}'.")       

def nop(n=1,H=0,F=0):
    ins = [(H << 29) | (F << 28)] * n
    asm(*ins)
    return ins

def bubble(n,flag=None):
    if flag is None:
        flag = getattr(asm,'bubble',None)
    if flag:
        nop(n)
    
def sel(rd,sbf,H=0,F=0):
    rd = rd.upper()
    RD, tmp, nRD = cnv_opd(rd, 'SEL', 'RD', 'R')
    if rd not in asm.core.SBF.keys():
        raise SyntaxError(f"'{rd}' is not a register sub-file.")
    sbf = sbf.upper()
    if sbf not in asm.core.SBF[rd]:
        raise SyntaxError(f"Invalid register '{sbf}' in sub-file {rd}.")
    RS = asm.core.sbfs[rd][sbf] + 0x80
    ins = (H << 29) | (F << 28) | ((RS >> 4) << 24) | (RD << 16) | ((RS & 0xF) << 12)
    asm(ins)
    return ins

def ldh(rd,imm,H=0,F=0,B=None):
    RD, tmp, nRD = cnv_opd(rd, 'LDH', 'RD', 'R')
    R0, iR0, nR0 = cnv_opd(imm, 'LDH', 'imm', 'A')
    ins = (H << 29) | (F << 28) | (RD << 16) | ((R0 >> 20) & 0xFFFF)
    asm(ins)
    if not (H or F):
        bubble(std_core.N_PLI,B)
    return ins

def ldl(rd,imm,H=0,F=0,B=None):
    RD, tmp, nRD = cnv_opd(rd, 'LDL', 'RD', 'R')
    R0, iR0, nR0 = cnv_opd(imm, 'LDL', 'imm', 'A')
    ins = 0x4000_0000 | (H << 29) | (F << 28) | (((R0 >> 16) & 0xF) << 24) | (RD << 16) | (R0 & 0xFFFF)
    asm(ins)
    if not (H or F):
        bubble(std_core.N_PLI,B)
    return ins

def ldi(rd,imm,H=0,F=0,B=None):
    RD, tmp, nRD = cnv_opd(rd, 'LDH', 'RD', 'R')
    R0, iR0, nR0 = cnv_opd(imm, 'LDH', 'imm', 'A')
    ins = [(H << 29) | (F << 28) | (RD << 16) | ((R0 >> 20) & 0xFFFF),
    0x4000_0000 | (H << 29) | (F << 28) | (((R0 >> 16) & 0xF) << 24) | (RD << 16) | (R0 & 0xFFFF)]
    asm(ins[0])
    R0, iR0, nR0 = cnv_opd(imm, 'LDH', 'imm', 'A')
    asm(ins[1])
    if not (H or F):
        bubble(std_core.N_PLI,B)
    return ins

def is_imm(val):
    if type(val) is str:
        return val[0] in ('#','-') or val[0].isdigit()
    return type(val) is int

def mov(rd,val,H=0,F=0,B=None):
    if callable(val):
        return val(rd,H=H,F=F,B=B)
    else:
        if is_imm(val):
            return ldi(rd,val,H=H,F=F,B=B)
        else:
            RD, tmp, nRD = cnv_opd(rd, 'MOV', 'RD', 'R')
            R1, iR1, nR1 = cnv_opd(val, 'MOV', "R1", "R")
            ins = ((2+nRD)<<30) | (H << 29) | (F << 28) | (asm.core.alu_opc['MOV'] << 24) | (RD << 16) | 0x8000 | (nR1 << 12) | R1
            asm(ins)
            if not (H or F):
                bubble(std_core.N_PLA,B)
            return ins
        
def alu(opc,rd,r0,r1,H=0,F=0,B=None):
    opc = opc.upper()
    if opc not in asm.core.alu_opc.keys():
        raise SyntaxError(f"Invalid opcode: '{opc}'.")
    RD, tmp, nRD = cnv_opd(rd, opc, 'RD', 'R')
    R0, iR0, nR0 = cnv_opd(r0, opc, "R0", "R" + asm.core.OPC[opc][0])
    R1, iR1, nR1 = cnv_opd(r1, opc, "R1", "R" + asm.core.OPC[opc][1])
    if iR0 == 1 and type(R0) is int and (0x000000_7F < R0 < 0xFFFFFF_C0):
        raise ValueError("R0 immediate out of bound.")
    elif iR0 == 0 and R0 > 0x3F:
        raise ValueError("R0 address out of bound.")
    if iR1 == 1 and type(R1) is int and (0x000000_7F < R1 < 0xFFFFFF_C0):
        raise ValueError("R1 immediate out of bound.")
    elif iR1 == 0 and R1 > 0x3F:
        raise ValueError("R1 address out of bound.")
    t0 = (R0 >> 6) & 1 if iR0 else nR0
    t1 = (R1 >> 6) & 1 if iR1 else nR1
    ins = ((2+nRD)<<30) | (H << 29) | (F << 28) | (asm.core.alu_opc[opc]<<24) | (RD<<16) | (((iR0<<3)+(iR1<<2)+(t0<<1)+t1)<<12) | ((R0&0x3F)<<6) | (R1&0x3F)
    asm(ins)
    if not (H or F):
        bubble(std_core.N_PLA,B)
    return ins

for k in base_core.OPC.keys()-['MOV']:
    globals()['and_' if k == 'AND' else k.lower()] = (lambda k:lambda rd,r0,r1,H=0,F=0,B=None:alu(k,rd,r0,r1,H,F,B))(k)

@multi(asm)                
def inline(prg):
    if type(prg) is str:
        prg = prg.splitlines()
    elif type(prg[0]) is int:
        prg = [source(ins) for ins in prg]
    for line in prg:
        instr = line.strip()
        if len(instr) == 0 or instr[0] == '%':
            continue
        if instr[0] == '#' and instr[-1] == ':':
            label(instr[1:-1])
        else:
            instr = instr.split()
            if len(instr) == 1:
                asm(int(instr[0],16))
            else:
                opc = instr[0].lower()
                h = int(instr[1].lower() == 'h') 
                f = int(instr[2].lower() == 'f')
                globals()['and_' if opc == 'and' else opc](*instr[3:],H=h,F=f,B=0)

@multi(asm)
def setup(core=C_STD, dnld=1):
    asm.core = core
    asm.dnld = dnld
    if dnld&1:
        asm.bubble = 1
        mov('trg','trg',B=0)
        timer()

@multi(asm)
def finish():
    dnld = getattr(asm,'dnld',1)
    if dnld&1:
        #rtlk_send(0)
        nop(2,H=1)
    
@multi(asm)
def timer(n=None,strict=True,wait=True):
    if n is None:
        if wait:
            sleep()
        mov('now','wck',B=0)
        mov('trg','trg',B=0)
        smk('err','nul','0x8.0',B=0)
    else:
        if is_imm(n):
            ldi('tr0',n,B=0)
            n = 'tr0'
        if wait:
            sleep()
        mov('now','wck',B=0)
        mov('tim',n,B=0)
        smk('err',('!' if strict else '')+'nul','0x4.0',B=0)

@multi(asm)
def sleep(n=None):
    if n is None:
        mov('nul','err',H=1)
    else:
        cyc = std_core.N_PLA + std_core.N_PLM
        if n <= cyc + 3:
            nop(n)
        else:
            n -= 3
            ldi('tr0',1-n//cyc,B=1)
            sne('ptr','ptr','tr0')
            add('tr0','tr0',1)
            nop(cyc + n % cyc - 2)

def wait_tx(chn_msk):
    nop()
    ldh('tr0',0xFFF00000,B=0)
    ldh('tr1',0xFFF00000,B=0)
    sne('ptr','ptr','tr0',B=0)
    ceq('!tr0','tr1','nul',B=0)
    ldi('tr1',chn_msk)
    and_('tr1','tr1','rtbc',B=0)
    nop(std_core.N_PLM)

@multi(asm)
def rtlk_send(dat, chn=0, wait=True):
    if wait:
        wait_tx(1 << chn)
    sel('rtda',f'C{chn:X}')
    mov('rtda',dat,B=0)

@multi(asm)
def set_rtcf(upl=None, bdr=None, hop=None, typ=None, bce=None, bcs=None):
    upl_msk, upl = (0, 0) if upl is None else (0xF, upl)
    bdr_msk, bdr = (0, 0) if bdr is None else (0xFF, bdr)
    hop_msk, hop = (0, 0) if hop is None else (0x7, hop)
    typ_msk, typ = (0, 0) if typ is None else (0x1, typ)
    bce_msk, bce = (0, 0) if bce is None else (0x7FFF, bce)
    bcs_msk, bcs = (0, 0) if bcs is None else (0x1, bcs)

    val = bit_concat((upl, 4), (bdr, 8), (hop, 3),
                     (typ, 1), (bce, 15), (bcs, 1))
    msk = bit_concat((upl_msk, 4), (bdr_msk, 8), (hop_msk, 3),
                     (typ_msk, 1), (bce_msk, 15), (bcs_msk, 1))
    ldi('tr0',val,B=0)
    ldi('tr1',msk)
    smk('rtcf','tr0','tr1')

@multi(asm)    
def jmp(dst,cond=None,met=True):
    if cond is None:
        if is_imm(dst):
            ldl('ptr',dst,F=1)
        else:
            mov('ptr',dst,F=1)
    else:
        if callable(cond):
            cond('tr1' if met else '!tr1')
            cond = 'tr1'
        elif not met:
            cond = '!'+cond
        if is_imm(dst):
            ldi('tr0',dst)
            dst = 'tr0'
        sne('ptr',dst,cond,F=1)

def block(tag=None,n=0):
    if n > 0:
        blk = table(tag=tag)
        asm.block = getattr(asm,'block',[]) + [blk]
        idx = len(asm.block) - 1
        for i in range(n):
            blk.append(f'{tag}_{idx}_{i}')
        return blk
    if tag is None:
        blk = asm.block.pop()
        if blk.tag == 'for':
            tr0 = blk.step
            if not (-0x40 <= tr0 < 0x40):
                ldi('tr0',tr0)
                tr0 = 'tr0'
            add(blk.rd,blk.rd,tr0,B=0)
        if blk.tag in ('forever','while','for'):
            jmp('#'+blk[0])
        if blk.tag != 'forever':
            blk[-1] = block(blk[-1])
        if blk.tag == 'elif':
            block()
        return blk
    elif type(tag) is str:
        pos = asm.label[tag]
        del asm.label[tag]
        tag = (pos[0],{i:asm[i] for i in pos[1:]})
    tag[0][0] = len(asm)
    for k,v in tag[1].items():
        asm[k] = v()
    return tag
        
@multi(asm)
def if_(cond):
    jmp('#'+block('if',2)[1],cond,False)

@multi(asm)
def elif_(cond):
    else_()
    jmp('#'+block('elif',2)[1],cond,False)

@multi(asm)
def else_():
    blk = asm.block[-1]
    jmp('#'+blk[0])
    block(blk.pop())

@multi(asm)    
def while_(cond=None):
    if cond is None:
        label(block('forever',1)[0])
    else:
        blk = block('while',2)
        label(blk[0])
        jmp('#'+blk[1],cond,False)

@multi(asm)
def for_(rd,rng):
    if type(rng) not in (tuple,list):
        rng = (rng,)
    if len(rng) == 1:
        start = 0
        stop = rng[0]
    else:
        start = rng[0]
        stop = rng[1]
    step = rng[2] if len(rng) > 2 else 1
    tr0 = stop
    if not (-0x40 <= tr0 < 0x40):
        ldi('tr0',tr0)
        tr0 = 'tr0'
    if step > 0:
        cond = lambda tr1,H=0,F=0,B=None:cls(tr1,rd,tr0,H=H,F=F,B=B)
    else:
        cond = lambda tr1,H=0,F=0,B=None:cls(tr1,tr0,rd,H=H,F=F,B=B)
    blk = block('for',2)
    blk.rd = rd
    blk.step = step
    mov(rd,start)
    label(blk[0])
    jmp('#'+blk[1],cond,False)

@multi(asm)    
def end():
    block()

@contextmanager
def If(cond):
    try:
        yield if_(cond)
    except Exception:
        raise
    else:
        asm.last_if = [block()]

@contextmanager
def Elif(cond):
    try:
        asm.block += asm.last_if
        yield elif_(cond)
    except Exception:
        raise
    else:
        asm.last_if.append(block())

@contextmanager
def Else():
    try:
        asm.block += asm.last_if
        yield else_()
    except Exception:
        raise
    else:
        block()
        asm.last_if = []

@contextmanager
def While(cond=None):
    try:
        yield while_(cond)
    except Exception:
        raise
    else:
        block()

@contextmanager
def For(rd, rng):
    try:
        yield for_(rd,rng)
    except Exception:
        raise
    else:
        block()

def frame(*args):
    if len(args) == 0:
        return getattr(asm, 'frame', (0,))
    asm.frame = args
    vars = [R[i] for i in range(sum(args))]
    return vars[0] if len(vars) == 1 else vars

@multi(asm)
def function(name, args=0, locals=0):
    label(name)
    for i in range(locals):
        mov('dts', f'R{(i+args):X}', B=0)
    return frame(args, locals)

@multi(asm)
def return_(*rets):
    for i in range(len(rets)):
        mov(f'R{(asm.core.n_gpr-1-i):X}', rets[i], B=0)
    size = sum(frame())
    for i in range(size):
        mov(f'R{(size-1-i):X}', 'dts')
    mov('ptr','rts',F=1)

@multi(asm)    
def call(name, *args):
    for i in range(len(args)):
        mov('dts', f'R{i:X}', B=0)
        mov(f'R{i:X}', args[i], B=0)
    add('rts','ptr',2,B=0)
    ldl('ptr','#'+name,F=1)
    return R[asm.core.n_gpr-1]

Return = return_
Call = call

@contextmanager
def Func(name, *regs):
    try:
        if len(regs) == 1:
            args = regs[0] + 1
            locals = 0
        else:
            args = regs[0]
            locals = regs[1] + 1 - regs[0]
        yield function(name,args,locals)
    except Exception:
        raise
    else:
        core = asm.core
        with asm:
            asm.core = core
            ins = mov('ptr','rts',F=1)
        if asm[-1] != ins:
            return_()

Set = mov
                    
class core_reg:
    def __init__(self, key=None):
        object.__setattr__(self,'_key',None if key is None else (f'R{key:X}' if type(key) is int else str(key).upper()))
    def __str__(self):
        return 'R' if self._key is None else self._key
    def __getattr__(self, key):
        return self[key]
    def __getitem__(self, key):
        if self._key is None:
            key = f'R{key:X}' if type(key) is int else str(key).upper()
            val = self.__dict__.get(key, None)
            if val is None:
                val = self.__class__(key)
                self.__dict__[key] = val
            return val
    def __setattr__(self, key, val):
        self[key] = val
    def __setitem__(self, key, val):
        if self._key is None:
            key = f'R{key:X}' if type(key) is int else str(key).upper()
            Set(key,val)
    def __eq__(self, other):
        return lambda rd,H=0,F=0,B=None:ceq(rd,self,other,H=H,F=F,B=B)
    def __ne__(self, other):
        return lambda rd,H=0,F=0,B=None:ceq('!'+rd,self,other,H=H,F=F,B=B)
    def __lt__(self, other):
        return lambda rd,H=0,F=0,B=None:cls(rd,self,other,H=H,F=F,B=B)
    def __gt__(self, other):
        return lambda rd,H=0,F=0,B=None:cls(rd,other,self,H=H,F=F,B=B)
    def __le__(self, other):
        return lambda rd,H=0,F=0,B=None:cls('!'+rd,other,self,H=H,F=F,B=B)
    def __ge__(self, other):
        return lambda rd,H=0,F=0,B=None:cls('!'+rd,self,other,H=H,F=F,B=B)
    def __add__(self, other):
        return lambda rd,H=0,F=0,B=None:add(rd,self,other,H=H,F=F,B=B)
    def __sub__(self, other):
        return lambda rd,H=0,F=0,B=None:add(rd,self,(-other) if type(other) is int else ('!'+str(other)),H=H,F=F,B=B)
    def __and__(self, other):
        return lambda rd,H=0,F=0,B=None:and_(rd,self,other,H=H,F=F,B=B)
    def __or__(self, other):
        return lambda rd,H=0,F=0,B=None:and_('!'+rd,'!'+str(self),'!'+str(other),H=H,F=F,B=B)
    def __xor__(self, other):
        return lambda rd,H=0,F=0,B=None:xor(rd,self,other,H=H,F=F,B=B)
    def __lshift__(self, other):
        return lambda rd,H=0,F=0,B=None:sla(rd,self,other,H=H,F=F,B=B)
    def __rshift__(self, other):
        return lambda rd,H=0,F=0,B=None:sla(rd,self,(-other) if type(other) is int else ('!'+str(other)),H=H,F=F,B=B)
    def __radd__(self, other):
        return self.__add__(other)
    def __rsub__(self, other):
        return lambda rd,H=0,F=0,B=None:add(rd,'!'+str(self),other,H=H,F=F,B=B)
    def __rand__(self, other):
        return self.__and__(other)
    def __ror__(self, other):
        return self.__or__(other)
    def __rxor__(self, other):
        return self.__xor__(other)
    def __rlshift__(self, other):
        return lambda rd,H=0,F=0,B=None:sla(rd,other,self,H=H,F=F,B=B)
    def __rrshift__(self, other):
        return lambda rd,H=0,F=0,B=None:sla(rd,other,'!'+str(self),H=H,F=F,B=B)
    def __neg__(self):
        return 0 - self
    def __imatmul__(self, other):
        R[self] = other
        return self
    def __iadd__(self, other):
        R[self] = self + other
        return self
    def __isub__(self, other):
        R[self] = self - other
        return self
    def __iand__(self, other):
        R[self] = self & other
        return self
    def __ior__(self, other):
        R[self] = self | other
        return self
    def __ixor__(self, other):
        R[self] = self ^ other
        return self
    def __ilshift__(self, other):
        R[self] = self << other
        return self
    def __irshift__(self, other):
        R[self] = self >> other
        return self
    
R = core_reg()

core_ctx = lambda core=asm.core:{k:R[k] for k in core.regs}|{k:globals()[k] for k in ('Set','If','Elif','Else','While','For','Return','Call','Func')}
core_regq = lambda core=asm.core:lambda s:s == 'R' or s in core.regs
core_domain = lambda core=asm.core,sub=True,dump=False:domain(core_ctx(core),core_regq(core),sub=sub,dump=dump)

@multi(asm)
def cache_write(adr, dat, typ='d'):
    mov(typ[0]+'ca',adr)
    mov(typ[0]+'cd',dat)

@multi(asm)
def cache_read(adr, typ='d'):
    mov(typ[0]+'ca',adr,B=1)
    nop(std_core.N_PLM)
    return typ[0]+'cd'

class core_cache:
    def __init__(self, typ, ptr=0):
        self.typ = typ
        self.ptr = ptr
    def __setitem__(self, key, val):
        cache_write(key, val, typ=self.typ)
    def __getitem__(self, key):
        return R[cache_read(key, typ=self.typ)]
    def __add__(self, ptr):
        return self.__class__(self.typ, self.ptr + ptr)
    def __call__(self, size):
        ptr = getattr(asm,self.typ,0)
        asm[self.typ] = ptr + size
        return self + ptr

DCH = core_cache('dat')
ICH = core_cache('ins')

def unpack_addr(addr):
    s = hex(addr).lstrip("0x")
    if "0" in s:
        raise RuntimeError(f"Invalid node address: 0x{addr:X}.")
    res = []
    for c in s:
        res += [int(c, 16)]
    return res

def pack_frame(hop, payload, typ=1):
    frm = bit_concat((C_BASE.MGW, 4), (hop, 3), (typ, 1), (payload, 32))
    return frm.to_bytes(5, "big")

def unpack_frame(frm):
    dat = int.from_bytes(frm, "big")
    mgw, hop, typ, pld = bit_split(dat, (4, 3, 1, 32))
    return pld, (mgw, hop, typ)

def send_to(addr, payloads, typ=1):
    """

    Send data / instruction frames to a node.

    Parameters
    ----------
    addr : integer
        Destination node address.
    payloads : list of integers
        Payloads of frames, each element corresponds to a frame.
    typ : integer
        Type of frames.
        0: data frame; 1: instruction frame;
        The default is 1.

    Returns
    -------
    buf : bytearray
        Byte stream of the frames.

    """
    hop = len(unpack_addr(addr))
    ln = len(payloads)
    buf = bytearray(ln * 5)
    head = bit_concat((C_BASE.MGW, 4), (hop, 3), (typ, 1)) << 32
    for i in range(ln):
        buf[i*5:(i+1)*5] = (head | payloads[i]).to_bytes(5, "big")
    return buf

def node_exec(addr, code):
    """

    Run a script on a node using configuration override.

    Parameters
    ----------
    addr : integer
        Destination node address.
    code : int array
        Machine code to be run.

    Returns
    -------
    buf : bytearray
        Byte stream of the script.
        
    """
    return send_to(addr, code, 1)

def suspend_node(addr):
    """

    Generate a script to suspend the destination node.

    Parameters
    ----------
    addr : integer
        Destination node address.

    Returns
    -------
    buf : bytearray
        Byte stream of the script.

    """
    with asm:
        return node_exec(addr,[smk('err','0x1.0','0x1.0')])

def resume_node(addr):
    """

    Generate a script to resume the destination node.

    Parameters
    ----------
    addr : integer
        Destination node address.

    Returns
    -------
    buf : bytearray
        Byte stream of the script.

    """
    with asm:
        return node_exec(addr,[smk('err','0x0.0','0x1.0')])

def reset_node(addr):
    """

    Generate a script to reset the Timer, TrigMgr & ErrorMgr of a node,
    and then jump to address 0.

    Parameters
    ----------
    addr : integer
        Destination node address.

    Returns
    -------
    buf : bytearray
        Byte stream of the script.

    """
    with asm:
        ldl('err',1,F=1)
        mov('trg','trg',H=1)
        mov('tim','nul',H=1)
        mov('ptr','nul',B=0)
        mov('ica','nul',B=0)
        mov('dca','nul',B=0)
        and_('!err','err','nul',B=0)
        return node_exec(addr, asm[:])

def download_to(addr, payloads, typ=1, start=0):
    """

    Generate a script to download data / program to the destination node.

    Parameters
    ----------
    addr : integer
        Destination node address.
    payloads : list of integers
        data / program in machine code.
    core : <CoreBase>, optional
        Destination core structure descriptor. The default is None.
    typ : integer, optional
        Type of payloads.
        0: data (to Data Cache); 1: program (to Instruction Cache);
        The default is 1.
    start : integer, optional
        Start address in the cache. The default is 0.

    Returns
    -------
    buf : bytearray
        Byte stream of the script.

    """
    cap = (asm.core.CAP_DCH, asm.core.CAP_ICH)[typ]
    ln = len(payloads)
    if ln > cap:
        tmp = ("data", "instruction")[typ]
        raise RuntimeError(f"Maximum {tmp} cache capacity exceeded.")
    scr = [0] * (ln * 3 + 1)
    rga = base_core.RSV.index(("DCA", "ICA")[typ])
    rgd = base_core.RSV.index(("DCD", "ICD")[typ])
    scr[0] = rga << 16
    for i in range(ln):
        scr[i*3+1] = 0x40000000 | (((i+start) & 0xF0000) << 8) | (rga << 16) | ((i+start) & 0xFFFF)
        scr[i*3+2] = (rgd << 16) | ((payloads[i] >> 20) & 0xFFF)
        scr[i*3+3] = 0x40000000 | ((payloads[i] & 0xF0000) << 8) | (rgd << 16) | (payloads[i] & 0xFFFF)
    return send_to(addr, scr, 1)

def reach_node(addr, ssp=True):
    """

    Construct a link between the computer and the destination node.
    So that the node can be controlled directly.

    Parameters
    ----------
    addr : integer
        Destination node address.
    ssp : bool, optional
        Whether to suspend the cores in the route. The default is True.

    Returns
    -------
    buf : bytearray
        Byte stream of the script.

    """
    nodes = unpack_addr(addr)
    hop = len(nodes)
    buf = bytearray()
    with asm:
        for i in range(hop):
            t_adr = addr >> ((hop - i) * 4)
            if ssp:
                buf += suspend_node(t_adr)
            rtcf = bit_concat((nodes[i], 4), (0, 12), ((1 << nodes[i]) + 1, 16))
            buf += node_exec(t_adr, ldi('rtcf',rtcf))
        rtcf = bit_concat((0, 12), (hop, 3), (0, 17))
        buf += node_exec(addr, ldi('rtcf',rtcf))
    return buf

class run_cfg:
    def __init__(self, intf, dst, core=None):
        self.intf = intf
        self.dst = dst
        self.core = core
        self.stat = 0

    def __call__(self, *args, **kwargs):
        if self.stat == 0:
            self.func = args[0]
            self.dnld = kwargs["dnld"] if "dnld" in kwargs else 1
            self.rply = kwargs["rply"] if "rply" in kwargs else 1
            self.dst  = kwargs["dst"] if "dst" in kwargs else self.dst
            self.tout = kwargs["tout"] if "tout" in kwargs else 1.0
            self.proc = kwargs["proc"] if "proc" in kwargs else None
            if type(self.func) is not table and callable(self.func):
                self.stat = 1
                return self
        self.stat = 0
        if not callable(self.func):
            flw = self.func
        else:
            with asm as flw:
                asm.intf = self.intf
                asm.tout = self.tout
                asm.proc = self.proc
                if type(self.func) is table:
                    nodes = getattr(self.func,'multi',None)
                    if nodes is None:
                        asm.dnld = self.dnld
                        asm(*self.func[:],**self.func.__dict__)
                    else:
                        asm.multi = nodes
                        for adr in nodes:
                            asm[str(adr)] = self.func[str(adr)].copy()
                            if getattr(asm[str(adr)],'dnld',None) is None:
                                asm[str(adr)].dnld = self.dnld
                else:
                    if self.core is not None:
                        asm.core = self.core
                    asm.dnld = self.dnld
                    res = self.func(*args, **kwargs)
                    if len(asm) == 0:
                        return res
                finish()
        with self.intf:
            res = []
            nodes = getattr(flw,'multi',None)
            if nodes is None:
                rply = getattr(flw,'rply',None)
                if rply is None:
                    rply = getattr(flw,'dnld',None)
                    rply = self.rply if rply is None else (rply&1)
                for adr in self.dst:
                    res += self.intf.run(flw.dnld, flw[:], adr, rply) or []
            else:
                flws = flw
                for i in range(len(self.dst)):
                    flw = flws[str(nodes[i])]
                    res += self.intf.run(flw.dnld, flw[:], self.dst[i], getattr(flw,'rply',(flw.dnld&1) and self.rply))
            proc = getattr(flw,'proc',self.proc)
            if proc is None:
                return None if len(res) == 0 else res
            else:
                return map(proc,res)
        
class assembler:
    def __init__(self, cfg=None, multi=None):
        self.cfg = cfg
        core = cfg and cfg.core or asm.core
        intf = cfg and cfg.intf or asm.intf
        with asm as self.asm:
            if multi is None:
                setup(core)
                asm.intf = intf
            else:
                nodes = []
                cores = []
                for i in multi:
                    if type(i) in (tuple,list):
                        nodes.append(str(i[0]))
                        cores.append(i[1])
                    else:
                        nodes.append(str(i))
                        cores.append(core)
                asm.multi = nodes
                for i in range(len(nodes)):
                    name = nodes[i]
                    setup[name](cores[i])
                    asm[name].intf = intf
                    self[name] = asm < asm[name]
                        
    def __getitem__(self, key):
        return getattr(self, str(key))
    
    def __setitem__(self, key, val):
        return setattr(self, str(key), val)
    
    def run(self, disa=False):
        if disa:
            multi = getattr(self.asm,'multi',None)
            if multi is None:
                print(source(self.asm))
            else:
                for i in multi:
                    print(i)
                    print(source(self.asm[str(i)]))
        if self.cfg and self.cfg.intf:
            return self.cfg(self.asm)
    
    def clear(self):
        multi = getattr(self.asm,'multi',None)
        if multi is None:
            self.asm.clear()
        else:
            for i in multi:
                self.asm[str(i)].clear()
        return self

    def __enter__(self):
        self.asm = asm <= self.asm

    def __exit__(self, exc_type, exc_value, traceback):
        self.asm = asm <= self.asm

    def __call__(self, *args, **kwargs):
        if len(args) > 0:
            multi = getattr(self.asm,'multi',None)
            if multi is None:
                with self:
                    args[0](*args[1:],**kwargs)
            elif callable(args[0]):
                for i in multi:
                    with self[i]:
                        args[0](i,*args[1:],**kwargs)
            elif len(args) > 1:
                env = self if args[0] is None else getattr(self,str(args[0]),None)
                if env is not None:
                    with env:
                        args[1](*args[2:],**kwargs)
        return self

def core_run(func):
    def wrap(*args, **kwargs):
        cfg = getattr(asm,'cfg',None)
        return func(*args, **kwargs) if cfg is None else cfg(func)(*args,**kwargs)
    return wrap

@core_run
def get_reg(rd,**kwargs):
    asm.dnld = kwargs.get('dnld',0)
    asm.rply = 1
    proc = kwargs.get('proc',None)
    if proc is not None:
        kwargs.pop('proc')
        asm.proc = lambda a, x: proc(x)
    rtlk_send(rd)

@core_run
def set_reg(rd,val):
    asm.dnld = 6
    mov(rd,val)

def get_err():
    err = get_reg('err',dnld=2)[0]
    print(f"ERR: 0x{err:X}")
    cfg = getattr(asm,'cfg',None)
    core = asm.core if cfg is None else cfg.core
    for i in range(len(core.T_ERR)):
        if err & (1<<i) != 0:
            print(f"{i:2}: {core.T_ERR[i]}")
    return err

def intf_run(func,sync=True):
    core = asm.core
    intf = getattr(asm, 'intf', None)
    if intf is None:
        cfg = getattr(asm, 'cfg', None)
        if cfg is None:
            return func
        intf = cfg.intf
        if cfg.core is not None:
            core = cfg.core
    oper = intf.oper.get((id(func)<<1)|sync,None)
    if oper is None:
        def cb(*args):
            if sync:
                with asm:
                    asm.core = core
                    adr = func(*args) or 0
                    intf.run(2,asm[:],adr)
                return True
            else:
                return func(*args)
        oper = len(intf.oper) + 1
        intf.oper[(id(func)<<1)|sync] = oper
        intf.oper[oper] = cb
    def wrap(*args,narg=None):
        if narg is None:
            narg = len(args)
        if hasattr(intf,'sn'):
            timer(300*1000)
            sleep()
        rtlk_send((narg<<20)|oper)
        if hasattr(intf,'sn'):
            timer(300*1000)
            sleep()
        for i in args:
            rtlk_send(i)
        if sync:
            smk('err','0x1.0','0x1.0')
    return wrap

intf_run_async = lambda func: intf_run(func,sync=False)

def segment(regs,vals,action=None):
    import base64
    tag = bytes([asm.core.regs[str(r)] for r in regs])
    if action is not None:
        tag += b'\x09' + id(action).to_bytes(8,'big')
    tag = base64.b64encode(tag).decode().rstrip('=')
    sub = getattr(asm,'sub',None)
    if sub is None:
        sub = {}
        asm.sub = sub
    n = len(regs)
    if sub.get(tag,None) is None:
        core = asm.core
        with asm:
            asm.core = core
            for i in range(n//7):
                for j in range(5):
                    add('dca','dca',j+1,B=0)
                add('dca','dca',5,B=0)
                add('dca','dca',5,B=0)
                for j in range(7):
                    mov(regs[7*i+j],'dcd',B=0)
            n %= 7
            if n == 0:
                if action is None:
                    mov('ptr','dcd',B=0)
                    add('dca','dca',1,B=0)
                    nop(7)
                elif callable(action):
                    action()
                else:
                    asm(*action)
            else:
                for j in range(n):
                    add('dca','dca',min(j+1,5),B=0)
                if action is None:
                    mov('ptr','dcd',B=0)
                    add('dca','dca',min(n+1,4),B=0)
                if n < 6:
                    nop((6 if action is None else 7)-n)
                for j in range(n):
                    mov(regs[-((n % 7)-j)],'dcd',B=0)
                if action is None:
                    nop()
                elif callable(action):
                    action()
                else:
                    asm(*action)
            sub[tag] = asm[:]
    dat = getattr(asm,'dat',None)
    if dat is None:
        dat = []
        asm.dat = dat
    if action is None and n > 0:
        dat += vals[:-n] + [None] + vals[-n:]
        return tag,len(dat)-1-n
    else:
        dat += vals + [None]
        return tag,len(dat)-1


import os,sys
PYODIDE = 'pyodide' in sys.modules
if PYODIDE:
    import js,pyodide
try:
    import veri
    veri.lib = os.path.join(os.path.dirname(os.path.realpath(__file__)),'..','..','..','hdl','rwg',sys.platform)
except:
    pass
try:    
    import serial
except:
    pass
try:
    import ftd3xx
except:
    pass
from . import asm,base_core,reach_node,suspend_node,node_exec,reset_node,resume_node,download_to,unpack_frame

class base_intf:
    def __init__(self):
        self.open_cnt = 0
        self.oper = {}
    
    def open(self):
        if self.open_cnt == 0:
            self.open_device()
        self.open_cnt += 1
    
    def close(self):
        self.open_cnt -= 1
        if self.open_cnt == 0:
            self.close_device()

    def __enter__(self):
        self.open()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def open_device(self):
        raise NotImplementedError()

    def close_device(self):
        raise NotImplementedError()

    def set_timeout(self, tout):
        raise NotImplementedError()

    def write(self, frm):
        raise NotImplementedError()

    def read(self):
        raise NotImplementedError()

    def alloc_read_buffer(self, cnt):
        raise NotImplementedError()
    
    def baud_reset(self):
        raise NotImplementedError()

    def reset(self):
        import time
        enter = self.open_cnt > 0
        while self.open_cnt > 0:
            self.__exit__(0, 0, 0)
        with self:
            cycleDevicePort = getattr(self.dev,'cycleDevicePort',None)
            if cycleDevicePort:
                cycleDevicePort()
        time.sleep(0.8)
        with self:
            self.baud_reset()
        time.sleep(0.05)
        if enter:
            self.__enter__()
            
    def run(self, dnld, prg, adr=0, cnt=0, tout=5, ssp=True):
        if type(prg) in (bytes,bytearray):
            buf = prg
        else:
            buf = reach_node(adr, ssp)
            if dnld & 1 == 0:
                if dnld & 2 == 0:
                    buf += suspend_node(adr)
                buf += node_exec(adr, prg)
                if not dnld & 4:
                    buf += resume_node(adr)
            else:
                buf += reset_node(adr)
                download = getattr(self,'download',None)
                if download:
                    download(adr, prg)
                else:
                    buf += download_to(adr, prg)
                buf += resume_node(adr)
            buf = bytes(buf)
        self.set_timeout(tout)
        if cnt > 0:
            self.alloc_read_buffer(cnt)
        elif cnt < 0:
            self.alloc_read_buffer(1)
        self.write(buf)
        if cnt > 0:
            return self.read()
        elif cnt == 0:
            return
        fin = True
        while True:
            try:
                pld = self.read()
            except TimeoutError:
                continue
            if fin:
                pld = pld[0]
                narg,oper = pld>>20,pld&0xfffff
                if oper == 0:
                    if narg == 0:
                        break
                else:
                    oper = self.oper[oper]
                self.set_timeout(tout)
                self.alloc_read_buffer(narg or 1)
                fin = False
            elif narg == 0:
                fin = oper(pld)
                self.set_timeout(tout)
                self.alloc_read_buffer(1)
            else:
                if oper == 0:
                    return pld
                else:
                    oper(*pld)
                    self.set_timeout(tout)
                    self.alloc_read_buffer(1)
                    fin = True

class ft601_intf(base_intf):
    def __init__(self, sn):
        self.sn = sn
        self.tout = 5.0
        super().__init__()
    
    def open_device(self):
        sn = bytes(self.sn, encoding="utf-8")
        self.dev = ftd3xx.create(sn, ftd3xx.FT_OPEN_BY_SERIAL_NUMBER)
        if self.dev is None:
            raise RuntimeError(f"Incorrect serial number: {self.sn}")

    def close_device(self):
        if self.dev is not None:
            self.dev.close()
        self.dev = None

    def set_timeout(self, tout):
        self.tout = tout
        self.dev.setPipeTimeout(0x02, int(tout * 1000))
        self.dev.setPipeTimeout(0x82, int(tout * 1000))

    def write(self, frm):
        self._write(self._frm_to_raw(frm))

    def read(self):
        dat = self._raw_to_frm(self._get_async_read())
        if len(dat) < 5*self.read_cnt:
            self.dev.abortPipe(0x02)
            self.dev.abortPipe(0x82)
            self.dev.resetDevicePort()
        payloads = []
        for i in range(self.read_cnt):
            frm = dat[5*i:5*(i+1)]
            if len(frm) < 5:
                raise RuntimeError(f"Data receive timed out, {len(payloads)} frames received.")
            pld, hdr = unpack_frame(frm)
            if hdr[0] == base_core.MGW_ERR:
                raise RuntimeError("Runtime error occured in RTMQ core.")
            elif hdr[0] != base_core.MGW:
                raise RuntimeError(f"Corrupted frame header: {hdr}.")
            elif hdr[2] == 0:
                payloads += [pld]
            else:
                return payloads
        return payloads

    def alloc_read_buffer(self, cnt):
        self._init_async_read(cnt * 8)
        self.read_cnt = cnt

    def baud_reset(self):
        self._write(b"\x00" * 8)

    def _write(self, dat):
        return self.dev.writePipe(0x02, dat, len(dat))

    def _sync_read(self, ln):
        return self.dev.readPipeEx(0x82, ln)[1]

    def _init_async_read(self, ln):
        self.ovlp = self.dev.initOverlapped()
        self.rd_dat = self.dev.asyncReadPipe(0x82, ln, self.ovlp)

    def _get_async_read(self):
        byt = self.dev.getOverlappedResult(self.ovlp)
        if self.dev.getLastError() != 0:
            print(f"getovlp: {ftd3xx.getStrError(self.dev.getLastError())}")
            self.dev.abortPipe(0x82)
            self.dev.resetDevicePort()
        self.dev.releaseOverlapped(self.ovlp)
        return self.rd_dat.raw[:byt]

    def _frm_to_raw(self, f):
        cnt = len(f) // 5
        tmp = [f[i*5:i*5+1]+b"\x00"*3+f[i*5+4:i*5:-1] for i in range(cnt)]
        return b"".join(tmp)
    
    def _raw_to_frm(self, r):
        cnt = len(r) // 8
        tmp = [r[i*8:i*8+1]+r[i*8+7:i*8+3:-1] for i in range(cnt)]
        return b"".join(tmp)

class uart_intf(base_intf):
    def __init__(self, port, baud):
        self.port = port
        self.baud = baud
        super().__init__()

    def open_device(self):
        self.dev = serial.Serial(self.port, self.baud)
        self.dev.reset_input_buffer()
        self.dev.reset_output_buffer()
        self.dev.stopbits = serial.STOPBITS_ONE

    def close_device(self):
        if self.dev is not None:
            self.dev.close()
        self.dev = None

    def set_timeout(self, tout):
        self.dev.timeout = tout
        self.dev.write_timeout = tout

    def write(self, frm):
        self.dev.write(frm)
    
    def read(self):
        payloads = []
        for i in range(self.read_cnt):
            frm = self.dev.read(5)
            if len(frm) < 5:
                raise RuntimeError(f"Data receive timed out, {len(payloads)} frames received.")
            pld, hdr = unpack_frame(frm)
            if hdr[0] == base_core.MGW_ERR:
                raise RuntimeError("Runtime error occured in RTMQ core.")
            elif hdr[0] != base_core.MGW:
                raise RuntimeError(f"Corrupted frame header: {hdr}.")
            elif hdr[2] == 0:
                payloads += [pld]
            else:
                return payloads
        return payloads

    def alloc_read_buffer(self, cnt):
        self.read_cnt = cnt

    def baud_reset(self):
        self.dev.write(b"\x00" * 5)

class sim_intf(base_intf):
    def __init__(self, top='top', io=None, trace=0):
        self.top = top.lower() if type(top) is str else top
        self.io = io or ((['FP_LED_MAIN']+[f'FP_LED_RWG{i+1}' for i in range(3)]) if self.top == 'top' else ['FP_LED_RWG1'])
        self.trace = trace
        super().__init__()
        
    def open_device(self):
        if callable(self.top):
            self.dev = self.top()
        else:
            if PYODIDE:
                import json
                js.eval(f'veri.${self.top}=veri.top("rwg.{self.top}",{json.dumps(self.io)},{self.trace})')
                self.dev = getattr(js.veri,f'${self.top}')
                self.dev.run()
            else:
                self.dev = veri.top(self.top,self.io,self.trace)
                self.dev.run()

    def close_device(self):
        if self.dev is not None:
            self.dev.close()
        self.dev = None

    def set_timeout(self, tout):
        self.tout = tout
        
    def write(self, frm):
        self.dev.write(frm)
    
    def read(self):
        payloads = []
        for i in range(self.read_cnt):
            frm = self.dev.read(5, self.tout)
            if PYODIDE:
                frm = frm.to_py()
            if len(frm) < 5:
                raise RuntimeError(f"Data receive timed out, {len(payloads)} frames received.")
            pld, hdr = unpack_frame(frm)
            if hdr[0] == base_core.MGW_ERR:
                raise RuntimeError("Runtime error occured in RTMQ core.")
            elif hdr[0] != base_core.MGW:
                raise RuntimeError(f"Corrupted frame header: {hdr}.")
            elif hdr[2] == 0:
                payloads += [pld]
            else:
                return payloads
        return payloads

    def alloc_read_buffer(self, cnt):
        self.read_cnt = cnt

    def baud_reset(self):
        self.dev.write(b'\x00' * 5)
    
    def download(self,addr,prg):
        dev = self.dev       
        if addr == 0:
            ich = dev.rwg__DOT__iRWG1__DOT__ICCH__DOT__MEM__BRA__0__KET____DOT__mem
            if ich is None:
                ich = dev.top__DOT__iMAIN__DOT__ICCH__DOT__MEM__BRA__0__KET____DOT__mem
        else:
            rwg = getattr(dev,f'__PVT__top__DOT__iRWG{addr}')
            ich = getattr(rwg,'__PVT__ICCH__DOT__MEM__BRA__0__KET____DOT__mem')
        for i in range(len(prg)):
            ich[i] = prg[i]

#intf = sim_intf('RWG')
#intf = ft601_intf('IONCxHYQ4')
#intf = uart_intf('COM3',2000000)
